/* Copyright (c) 2006,2007,2008,2009 Polybus Systems Corp., Westford, MA, USA  */

/* All rights reserved. */

/* Redistribution and use in source and binary forms, with or without modification, */
/* are permitted provided that the following conditions are met: */

/* * Redistributions of source code must retain the above copyright notice, this */
/* list of conditions and the following disclaimer. */
/* * Redistributions in binary form must reproduce the above copyright notice, */
/* this list of conditions and the following disclaimer in the documentation */
/* and/or other materials provided with the distribution. */
/* * Neither the name of the Polybus Systems Corp nor the names of its contributors */
/* may be used to endorse or promote products derived from this software */
/* without specific prior written permission. */

/* THIS SOFTWARE IS PROVIDED BY POLYBUS AND CONTRIBUTORS ''AS IS'' AND ANY */
/* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE */
/* DISCLAIMED. IN NO EVENT SHALL POLYBUS OR CONTRIBUTORS BE LIABLE FOR ANY */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES */
/* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON */
/* ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT */
/* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

#include "sys_basher.h"

long fixedTest(SInt64 *base,SInt64 limit,SInt32 num,SInt32 test_num,result rslt[NUM_FIXED],SInt32 vnum)
{
  SInt64 *sindx,i,j,err_bits[2]={0,0},diff=0,words,value,tval[4],x,y;
  SInt32 lsb;
  char string[256],xstring[512];

  words = limit/8LL;
  x = test_num & 0x07LL;
  y = test_num >> 3;
  value = x * 0x1111111111111111LL;

  rslt[test_num].size = limit;
  printf("Fixed Memory Test %lld Bytes\n",limit);

  if(y)
    {
      tval[0] = value;
      tval[1] = ~value;
      tval[2] = value;
      tval[3] = ~value;
    }
  else
    {
      tval[0] = value;
      tval[1] = value;
      tval[2] = ~value;
      tval[3] = ~value;
    }

  if(base)
    {
      sindx = base;
      diff = 0;
      err_bits[0] = 0;
      err_bits[1] = 0;
      for(i=0;i<words;i+=4LL)
	{
	  *sindx++ = tval[0];
	  *sindx++ = tval[1];
	  *sindx++ = tval[2];
	  *sindx++ = tval[3];
	}
      for(j=0;j<2;j++)
	{
	  sindx = base;
	  for(i=0;i<words;i++)
	    {
	      value = *sindx;
	      *sindx++ = ~value;
	    }

	  for(i=0;i<words;i++)
	    {
	      sindx--;
	      value = *sindx;
	      *sindx = ~value;
	    }
	}
      lsb = 0x1L;
      for(i=0;i<words;i+=4LL)
	{
	  for(j=0;j<4;j++)
	    {
	      value = *sindx++;
	      diff = value ^ tval[j];
	      lsb ^= 0x1L;
	      err_bits[lsb] |= diff;
	    }
	}

      sprintf(rslt[test_num].name,"Mem Pattern: %1.1llX %1.1llX %1.1llX %1.1llX",
	      tval[0] & 0x0FL,
	      tval[1] & 0x0FL,
	      tval[2] & 0x0FL,
	      tval[3] & 0x0FL);
      sprintf(string,"Thread %ld, Patterns: %16.16llX %16.16llX %16.16llX %16.16llX",(long)vnum,tval[0],tval[1],tval[2],tval[3]);
      rslt[test_num].runs = rslt[test_num].runs + 1;
      if((err_bits[0] == 0) && (err_bits[1] == 0))
	{
	  sprintf(xstring,"PASS, %s\n",string);
	  printTranscript(xstring,TRUE);
	  return(1L);
	}
      else
	{
	  sprintf(xstring,"FAIL, %s, Error Mask:%16.16llX %16.16llX\n",string,err_bits[0],err_bits[1]);
	  printTranscript(xstring,TRUE);
	  rslt[test_num].errors = rslt[test_num].errors + 1;
	  return(0L);
	}
    }
  return(0L);
}

long addrEqDataTest(SInt64 *base,SInt64 limit,SInt32 num,SInt32 test_num,result rslt[NUM_ADDREQ],SInt32 vnum)
{
  SInt64 *sindx,i,err_bits[2]={0,0},diff=0,words,x,y,value,exp;
  SInt32 lsb;
  SInt64 k,com;
  char string[256],xstring[512];

  words = limit/8LL;

  rslt[test_num].size = limit;
  printf("A==D Memory Test %lld Bytes\n",limit);

  if(base)
    {
      sindx = base;
      for(i=0;i<words;i++)
	{
	  x = i * 0x0000000100000001LL;
	  y = x >> (64 - test_num);
	  x = x << test_num;
	  exp = x | y;
	  *sindx++ = exp;
	}
      com = 0x1111111111111111LL;
      for(k=0;k<2;k++)
	{
	  sindx = base;
	  for(i=0;i<words;i++)
	    {
	      value = *sindx;
	      *sindx++ = value ^ com;
	    }

	  com = com << 1LL;
	  for(i=0;i<words;i++)
	    {
	      sindx--;
	      value = *sindx;
	      *sindx = value ^ com;
	    }
	  com = com << 1LL;
	}
      sindx = base;
      lsb = 1;
      err_bits[0] = 0;
      err_bits[1] = 0;
      for(i=0;i<words;i++)
	{
	  x = i * 0x0000000100000001LL;
	  y = x >> (64 - test_num);
	  x = x << test_num;
	  exp = (x | y);
	  value = *sindx;
	  value = ~value;
	  *sindx++ = value;
	  diff = exp ^ value;
	  lsb ^= 0x1L;
	  err_bits[lsb] |= diff;
	}
      sindx = base;
      lsb = 1;
      for(i=0;i<words;i++)
	{
	  x = i * 0x0000000100000001LL;
	  y = x >> (64 - test_num);
	  x = x << test_num;
	  exp = (x | y);
	  value = *sindx++;
	  diff = exp ^ value;
	  lsb ^= 0x1L;
	  err_bits[lsb] |= diff;
	}
      sprintf(rslt[test_num].name,"Mem: Data = Addr << %ld",(long)test_num);
      sprintf(string,"Thread %ld, Pattern: Data = Addr << %ld",(long)vnum,(long)test_num);
      rslt[test_num].runs = rslt[test_num].runs + 1;
      if((err_bits[0] == 0) && (err_bits[1] == 0))
	{
	  sprintf(xstring,"PASS, %s\n",string);
	  printTranscript(xstring,TRUE);
	  return(1L);
	}
      else
	{
	  sprintf(xstring,"FAIL, %s, Error Mask:%16.16llX %16.16llX\n",string,err_bits[0],err_bits[1]);
	  printTranscript(xstring,TRUE);
	  rslt[test_num].errors = rslt[test_num].errors + 1;
	  return(0L);
	}
	  
    }
  return(0L);

}



long bankTest(SInt64 *base,SInt64 limit,SInt32 num,SInt32 test_num,result rslt[NUM_BANK],SInt32 vnum)
{
  SInt64 *xindx,i,j,err_bits[2]={0,0},diff=0,words,value,exp,*indx[16],offset;
  SInt32 lsb;
  char string[256],xstring[512];
  words = limit/8LL;

  offset = words/test_num;

  rslt[test_num].size = limit;
  printf("Bank Memory Test %lld Bytes\n",limit);

  if(base)
    {
      j = 0;
      for(i=0;i<test_num;i++)
	{
	  indx[i] = &base[j];
	  j+= offset;
	}

      for(i=0;i<offset;i++)
	{
	  value = i * 0x0123456789ABCDEFLL;
	  for(j=0;j<test_num;j++)
	    {
	      xindx = indx[j];
	      *xindx++ = value;
	      indx[j] = xindx;
	      value += 0x1111111111111111LL;
	    }
	}

      j = 0;
      for(i=0;i<test_num;i++)
	{
	  indx[i] = &base[j];
	  j+= offset;
	}
      for(i=0;i<offset;i++)
	{
	  for(j=0;j<test_num;j++)
	    {
	      xindx = indx[j];
	      value = *xindx;
	      *xindx++ = ~value;
	      indx[j] = xindx;
	    }
	}
      for(i=0;i<offset;i++)
	{
	  for(j=0;j<test_num;j++)
	    {
	      xindx = indx[j];
	      xindx--;
	      value = *xindx;
	      *xindx = ~value;
	      indx[j] = xindx;
	    }
	}
      j = 0;
      for(i=0;i<test_num;i++)
	{
	  indx[i] = &base[j];
	  j+= offset;
	}
      lsb = 1L;
      for(i=0;i<offset;i++)
	{
	  exp = i * 0x0123456789ABCDEFLL;
	  for(j=0;j<test_num;j++)
	    {
	      xindx = indx[j];
	      value = *xindx++;
	      indx[j] = xindx;
	      diff = exp ^ value;
	      lsb ^= 0x1L;
	      err_bits[lsb] |= diff;
	      exp += 0x1111111111111111LL;
	    }
	}
    }
  sprintf(rslt[test_num].name,"Multibank: %ld banks",(long)test_num);
  sprintf(string,"Thread %ld, Multi Bank Test (%ld banks)",(long)vnum,(long)test_num);
  rslt[test_num].runs = rslt[test_num].runs + 1;
  if((err_bits[0] == 0) && (err_bits[1] == 0))
    {
      sprintf(xstring,"PASS, %s\n",string);
      printTranscript(xstring,TRUE);
      return(1L);
    }
  else
    {
      sprintf(xstring,"FAIL, %s, Error Mask:%16.16llX %16.16llX\n",string,err_bits[0],err_bits[1]);
      printTranscript(xstring,TRUE);
      rslt[test_num].errors = rslt[test_num].errors + 1;
      return(0L);
    }
}
      

long swapTest(SInt64 *base,SInt64 limit,SInt32 num,result *rslt,SInt32 vnum)
{
  SInt64 *xindx,i,err_bits[2]={0,0},words,x,value;
  SInt32 lsb;
  char string[256],xstring[512];

  words = limit/8LL;

  rslt->size = limit;

  if(base)
    {
      xindx = base;
      for(i=0;i<words;i++)
	{
	  value = i * 0xFEDCBA9876543211LL;
	  *xindx++ = value;
	}

      xindx = base;
      lsb = 1L;
      for(i=0;i<words;i++)
	{
	  value = i * 0xFEDCBA9876543211LL;
	  x = *xindx++;
	  lsb ^= 0x1L;
	  err_bits[lsb] |= value ^ x; 
	}

    }
  sprintf(string,"Thread %ld, Swap test",(long)vnum);
  sprintf(rslt->name,"Swap");
  rslt->runs = rslt->runs + 1;
  if((err_bits[0] == 0) && (err_bits[1] == 0))
    {
      sprintf(xstring,"PASS, %s\n",string);
      printTranscript(xstring,TRUE);
      return(1L);
    }
  else
    {
      sprintf(xstring,"FAIL, %s, Error Mask:%16.16llX %16.16llX\n",string,err_bits[0],err_bits[1]);
      printTranscript(xstring,TRUE);
      rslt->errors = rslt->errors + 1;
      return(0L);
    }
}




long memBandwidth(SInt64 *base,SInt64 limit,SInt32 num,SInt32 num_threads,result rslt[2][MAX_SIZES],SInt32 vnum)
{
  SInt64 *sindx,i,j,length,max,acc,loops,exp,bytes;
  SInt64 k,errors[64],total_errors;
  double start,end,wusecs[64],rusecs[64],bwidth[4],len;
  char string[1024],xstring[1024];
  char *test_names[]={"Read","Write","PASS/FAIL",0L};

  max = limit & 0xFFFFFFFFFFFF0000LL;
  max /= sizeof(SInt64);
  length = 0x80L;
  k = 0;
  sprintf(string,"\nThread %ld of %ld: Running Memory Bandwidth Tests\n",(long)vnum,(long)num_threads);
  printTranscript(xstring,FALSE);
  while(length < max)
    {
      loops = max/length;
      start = getTime();


      for(i=0;i<loops;i++)
	{
	  sindx = base;
	  for(j=0;j<length;j+=4)
	    {
	      *sindx++ = 1;
	      *sindx++ = 1;
	      *sindx++ = 1;
	      *sindx++ = 1;
	    }
	}
      end = getTime();
      wusecs[k] = end - start;
      start = getTime();
      acc = 0;
      for(i=0;i<loops;i++)
	{
	  sindx = base;
	  for(j=0;j<length;j+=4)
	    {
	      acc += *sindx++;
	      acc += *sindx++;
	      acc += *sindx++;
	      acc += *sindx++;
	    }
	}
      end = getTime();
      rusecs[k] = end - start;
      exp = loops * length;
      errors[k] = acc != exp;
      
      k++;
      length = length << 1;
    }
  length = 0x80L;
  displayTitles("Array Size",test_names,3L);
  total_errors = 0;
  for(i=0;i<k;i++)
    { 
      bytes = length * sizeof(SInt64);
      iscale(string,"B",bytes);
      loops = max/length;
      len = bytes;
      len *= loops;
      bwidth[0] = len/rusecs[i];
      bwidth[1] = len/wusecs[i];
      displayStatusLine(string,"B/S",bwidth,2L,errors[i]);
      length = length << 1;
      total_errors += errors[i];
      updateBandwidth(bwidth[0]/1024.0,rusecs[i],len/1024.0,&rslt[0][i]);
      updateBandwidth(bwidth[1]/1024.0,wusecs[i],len/1024.0,&rslt[1][i]);
      sprintf(rslt[0][i].name,"Memory Read Bandwidth");
      sprintf(rslt[1][i].name,"Memory Write Bandwidth");
      rslt[0][i].errors += errors[i];
      rslt[1][i].errors += errors[i];
      rslt[0][i].runs = rslt[0][i].runs + 1;
      rslt[1][i].runs = rslt[1][i].runs + 1;
      rslt[0][i].size = bytes;
      rslt[1][i].size = bytes;
      
    }
  sprintf(string,"Thread %ld, Memory Bandwidth Test",(long)vnum);
  return(total_errors == 0L);
}

void getLFSR(UInt64 data[16])
{
  UInt64 acc,lsbs;
  SInt32 i;
  acc = data[15];
  for(i=0;i<16;i++)
    {
      lsbs = acc << 61;
      acc = acc ^ (acc << 2LL);
      acc = acc >> 3LL;
      acc |= lsbs;
      data[i] = acc ^ 0x01LL;
    }
  
}

long lfsrTest(SInt64 *base,SInt64 limit,SInt32 num,SInt32 test_num,result rslt[NUM_LFSR],SInt32 vnum,UInt64 *seed, Boolean testRefresh)
{
  SInt64 *sindx,*hindx,*top,i,j,k,l,err_bits[2]={0,0},diff=0,words,value,high;
  SInt32 lsb;
  UInt64 lfsr[16],init_seed=*seed;
  char string[256],xstring[512];


  words = limit/8LL;

  rslt[test_num].size = limit;
  if(testRefresh)
    {
      printf("Memory Refresh Test %lld Bytes\n",limit);
    }
  else
    {
      printf("Random Number Memory Test %lld Bytes\n",limit);
    }

  if(base)
    {
      for(l=0;l<2;l++)
	{
	  sindx = base;
	  for(i=0;i<16;i++)
	    {
	      lfsr[i] = init_seed;
	    }
	  for(i=0;i<words;i+=16)
	    {
	      getLFSR(lfsr);
	      for(j=0;j<16;j++)
		{
		  *sindx++ = lfsr[j];
		}
	    }
	  top = sindx;
	  if(testRefresh)
	    {
	      delay(30L);
	    }
	  // Swap all locations from top to bottom and then do it again
	  for(k=0;i<2;k++)
	    {
	      hindx = top;
	      sindx = base;
	      for(i=0;i<words;i++)
		{
		  hindx--;
		  value = *sindx;
		  high = *hindx;
		  *sindx++ = high;
		  *hindx = value;
		}
	    }
	  for(i=0;i<16;i++)
	    {
	      lfsr[i] = init_seed;
	    }

	  err_bits[0] = 0;
	  err_bits[1] = 0;
	  sindx = base;
	  lsb = 1L;
	  for(i=0;i<words;i+=16)
	    {
	      getLFSR(lfsr);
	      for(j=0;j<16;j++)
		{
		  value = *sindx++;
		  diff = value ^ lfsr[j];
		  lsb ^= 0x1L;
		  err_bits[lsb] |= diff;
		}
	    }
	  init_seed = lfsr[15];
	}
      if(testRefresh)
	{
	  sprintf(rslt[test_num].name,"Mem: Refresh");
	  sprintf(string,"Thread %ld, Memory Refresh",(long)vnum);
	}
      else
	{
	  sprintf(rslt[test_num].name,"Mem: Random Data [Seed %16.16llX]",*seed);
	  sprintf(string,"Thread %ld, Random Data [Seed %16.16llX]",(long)vnum,*seed);
	}
      rslt[test_num].runs = rslt[test_num].runs + 1;
      *seed = init_seed;
      //      printf("Last LSFR: %16.16llX\n",init_seed);
      if((err_bits[0] == 0) && (err_bits[1] == 0))
	{
	  sprintf(xstring,"PASS, %s\n",string);
	  printTranscript(xstring,TRUE);
	  return(1L);
	}
      else
	{
	  sprintf(xstring,"FAIL, %s, Error Mask:%16.16llX %16.16llX\n",string,err_bits[0],err_bits[1]);
	  printTranscript(xstring,TRUE);
	  rslt[test_num].errors = rslt[test_num].errors + 1;
	  return(0L);
	}
	  
    }
  return(0L);

}

long walkTest(SInt64 *base,SInt64 limit,SInt32 num,SInt32 test_num,result rslt[NUM_ADDREQ],SInt32 vnum)
{
  SInt64 *sindx,i,j,err_bits[2]={0,0},diff=0,words,value,fvalue,exp,shf,nexp,nvalue;
  SInt32 lsb;
  char string[256],xstring[512];

  words = limit/8LL;

  rslt[test_num].size = limit;
  printf("Walking One/Zero Memory Test %lld Bytes\n",limit);

  shf = test_num & 0x7LL;
  fvalue = 0x0101010101010101LL << shf;
  if(base)
    {
      sindx = base;
      for(i=0;i<words;i+=8)
	{
	  value = fvalue;
	  for(j=0;j<8;j++)
	    {
	      *sindx++ = value;
	      value <<= 1L;
	      if(value == 0x0101010101010100LL) value = 0x0101010101010101LL;
	    }
	}
      err_bits[0] = 0;
      err_bits[1] = 0;
      sindx = base;
      lsb = 1L;
      for(i=0;i<words;i+=8)
	{
	  exp = fvalue;
	  for(j=0;j<8;j++)
	    {
	      value = *sindx++;
	      lsb ^= 0x1L;
	      diff = exp ^ value;
	      err_bits[lsb] |= diff;
	      exp <<= 1L;
	      if(exp == 0x0101010101010100LL) exp = 0x0101010101010101LL;
	    }
	}

      sindx = base;
      for(i=0;i<words;i++)
	{
	  value = *sindx;
	  nvalue = ~value;
	  *sindx++ = nvalue;
	}

      sindx = base;
      lsb = 1L;
      for(i=0;i<words;i+=8)
	{
	  exp = fvalue;
	  for(j=0;j<8;j++)
	    {
	      value = *sindx++;
	      nexp = ~exp;
	      lsb ^= 0x1L;
	      diff = nexp ^ value;
	      err_bits[lsb] |= diff;
	      exp <<= 1L;
	      if(exp == 0x0101010101010100LL) exp = 0x0101010101010101LL;
	    }
	}

      sprintf(rslt[test_num].name,"Mem: Walking One/Zero %16.16llX",fvalue);
      sprintf(string,"Thread %ld, Pattern: Walking One/Zero %16.16llX",(long)vnum,fvalue);
      rslt[test_num].runs = rslt[test_num].runs + 1;
      if((err_bits[0] == 0) && (err_bits[1] == 0))
	{
	  sprintf(xstring,"PASS, %s\n",string);
	  printTranscript(xstring,TRUE);
	  return(1L);
	}
      else
	{
	  sprintf(xstring,"FAIL, %s, Error Mask:%16.16llX %16.16llX\n",string,err_bits[0],err_bits[1]);
	  printTranscript(xstring,TRUE);
	  rslt[test_num].errors = rslt[test_num].errors + 1;
	  return(0L);
	}
	  
    }
  return(0L);

}

void getBitRev(UInt64 array[1024],UInt64 start,SInt32 width)
{
  SInt32 i,j;
  UInt64 acc,rev;

  for(i=0;i<1024;i++)
    {
      acc = start + i;
      rev = acc & 0x1L;
      for(j=0;j<width;j++)
	{
	  acc >>= 1;
	  rev <<= 1;
	  rev |= acc & 0x1LL;
	}
      array[i] = rev;
    }
}

SInt32 limitToWidth(UInt64 limit)
{
  SInt32 width=-1;
  UInt64 acc=1L;
  while(acc < limit)
    {
      acc <<= 1;
      width++;
    }
  if(width < 0) width = 0;
  return(width);
}

long revAddressTest(SInt64 *base,SInt64 limit,SInt32 num,SInt32 test_num,result rslt[NUM_ADDREQ],SInt32 vnum)
{
  SInt64 i,j,err_bits[2]={0,0},diff=0,words,value,exp,nexp,lsb,max;
  SInt32 width=0;
  UInt64 index[1024];
  char string[256],xstring[512];
  
  words = limit/8LL;
  width = limitToWidth(words);
  width--;
  if(width < 0) width = 0;
  max = 0x2LL << width;

  rslt[test_num].size = limit;
  printf("Bit Reverse Address Bottom Up Memory Test 0x%llx Words\n",max);

  if(base)
    {
      value = 0x123456789ABCDEFLL;
      for(i=0;i<max;i+=1024)
	{
	  getBitRev(index,i,width);
	  for(j=0;j<1024;j++)
	    {
	      base[index[j]] = value;
	      value += 0x0101010101010101LL;
	    }
	}
      err_bits[0] = 0;
      err_bits[1] = 0;

      exp = 0x123456789ABCDEFLL;
      for(i=0;i<max;i+=1024)
	{
	  getBitRev(index,i,width);
	  for(j=0;j<1024;j++)
	    {
	      value = base[index[j]];
	      lsb = index[j] & 0x1LL;
	      base[index[j]] = ~exp;
	      diff = exp ^ value;
	      err_bits[lsb] |= diff;
	      exp += 0x0101010101010101LL;
	    }
	}

      exp = 0x123456789ABCDEFLL;
      for(i=0;i<max;i+=1024)
	{
	  getBitRev(index,i,width);
	  for(j=0;j<1024;j++)
	    {
	      value = base[index[j]];
	      lsb = index[j] & 0x1LL;
	      nexp = ~exp;
	      diff = nexp ^ value;
	      err_bits[lsb] |= diff;
	      exp += 0x0101010101010101LL;
	    }
	}

      width += 3L;
      sprintf(rslt[test_num].name,"Mem: Bit Reverse Address Bottom Up");
      sprintf(string,"Thread %ld, Bit Reverse Address Bottom Up",(long)vnum);
      rslt[test_num].runs = rslt[test_num].runs + 1;
      if((err_bits[0] == 0) && (err_bits[1] == 0))
	{
	  sprintf(xstring,"PASS, %s\n",string);
	  printTranscript(xstring,TRUE);
	  return(1L);
	}
      else
	{
	  sprintf(xstring,"FAIL, %s, Error Mask:%16.16llX %16.16llX\n",string,err_bits[0],err_bits[1]);
	  printTranscript(xstring,TRUE);
	  rslt[test_num].errors = rslt[test_num].errors + 1;
	  return(0L);
	}
	  
    }
  return(0L);

}

long revAddressTopDownTest(SInt64 *base,SInt64 limit,SInt32 num,SInt32 test_num,result rslt[NUM_ADDREQ],SInt32 vnum)
{
  SInt64 i,j,err_bits[2]={0,0},diff=0,words,value,exp,nexp,lsb,top,max;
  SInt32 width=0;
  UInt64 index[1024];
  char string[256],xstring[512];
  
  words = limit/8LL;
  width = limitToWidth(words);
  width--;
  if(width < 0) width = 0;
  max = 0x2LL << width;
  top = words - 1;
  rslt[test_num].size = limit;
  printf("Bit Reverse Address Memory Top Down Test 0x%llx Words\n",max);

  if(base)
    {
      value = 0x123456789ABCDEFLL;
      for(i=0;i<max;i+=1024)
	{
	  getBitRev(index,i,width);
	  for(j=1023;j>=0;j--)
	    {
	      base[top - index[j]] = value;
	      value += 0x0101010101010101LL;
	    }
	}
      err_bits[0] = 0;
      err_bits[1] = 0;

      exp = 0x123456789ABCDEFLL;
      for(i=0;i<max;i+=1024)
	{
	  getBitRev(index,i,width);
	  for(j=1023;j>=0;j--)
	    {
	      value = base[top - index[j]];
	      lsb = index[j] & 0x1LL;
	      base[top - index[j]] = ~exp;
	      diff = exp ^ value;
	      err_bits[lsb] |= diff;
	      exp += 0x0101010101010101LL;
	    }
	}

      exp = 0x123456789ABCDEFLL;
      for(i=0;i<max;i+=1024)
	{
	  getBitRev(index,i,width);
	  for(j=1023;j>=0;j--)
	    {
	      value = base[top - index[j]];
	      lsb = index[j] & 0x1LL;
	      nexp = ~exp;
	      diff = nexp ^ value;
	      err_bits[lsb] |= diff;
	      exp += 0x0101010101010101LL;
	    }
	}

      width += 3L;
      sprintf(rslt[test_num].name,"Mem: Bit Reverse Address Top Down");
      sprintf(string,"Thread %ld, Bit Reverse Address Top Down",(long)vnum);
      rslt[test_num].runs = rslt[test_num].runs + 1;
      if((err_bits[0] == 0) && (err_bits[1] == 0))
	{
	  sprintf(xstring,"PASS, %s\n",string);
	  printTranscript(xstring,TRUE);
	  return(1L);
	}
      else
	{
	  sprintf(xstring,"FAIL, %s, Error Mask:%16.16llX %16.16llX\n",string,err_bits[0],err_bits[1]);
	  printTranscript(xstring,TRUE);
	  rslt[test_num].errors = rslt[test_num].errors + 1;
	  return(0L);
	}
	  
    }
  return(0L);

}
