/* Copyright (c) 2006,2007,2008,2009 Polybus Systems Corp., Westford, MA, USA  */

/* All rights reserved. */

/* Redistribution and use in source and binary forms, with or without modification, */
/* are permitted provided that the following conditions are met: */

/* * Redistributions of source code must retain the above copyright notice, this */
/* list of conditions and the following disclaimer. */
/* * Redistributions in binary form must reproduce the above copyright notice, */
/* this list of conditions and the following disclaimer in the documentation */
/* and/or other materials provided with the distribution. */
/* * Neither the name of the Polybus Systems Corp nor the names of its contributors */
/* may be used to endorse or promote products derived from this software */
/* without specific prior written permission. */

/* THIS SOFTWARE IS PROVIDED BY POLYBUS AND CONTRIBUTORS ''AS IS'' AND ANY */
/* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED */
/* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE */
/* DISCLAIMED. IN NO EVENT SHALL POLYBUS OR CONTRIBUTORS BE LIABLE FOR ANY */
/* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES */
/* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON */
/* ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT */
/* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. */

typedef unsigned char Boolean;

#define TRUE 1
#define FALSE 0

#define MAX_CORES 32
#define MAX_THREADS ((MAX_CORES +1) * (MAX_CORES/2)) + 1
#define MAX_DIRS 32
#define MAX_FILE_SIZES 28
#define MAX_SIZES 32

#define NUM_FIXED 16
#define NUM_BANK 16
#define NUM_ADDREQ 32
#define NUM_LFSR 8
#define NUM_WALK 8
#define NUM_REV_ADDR 2

typedef int16_t SInt16;
typedef int32_t SInt32;
typedef long long int SInt64;

typedef u_int16_t UInt16;
typedef u_int32_t UInt32;
typedef long long unsigned int UInt64;


enum{
  WALK_TST0=0,
  WALK_TST1,
  WALK_TST2,
  WALK_TST3,
  WALK_TST4,
  WALK_TST5,
  WALK_TST6,
  WALK_TST7,
  FIXED_MEM0,
  FIXED_MEM1,
  FIXED_MEM2,
  FIXED_MEM3,
  FIXED_MEM4,
  FIXED_MEM5,
  FIXED_MEM6,
  FIXED_MEM7,
  FIXED_MEM8,
  FIXED_MEM9,
  FIXED_MEM10,
  FIXED_MEM11,
  FIXED_MEM12,
  FIXED_MEM13,
  FIXED_MEM14,
  FIXED_MEM15,
  ADDR_EQ0,
  ADDR_EQ1,
  ADDR_EQ2,
  ADDR_EQ3,
  ADDR_EQ4,
  ADDR_EQ5,
  ADDR_EQ6,
  ADDR_EQ7,
  ADDR_EQ8,
  ADDR_EQ9,
  ADDR_EQ10,
  ADDR_EQ11,
  ADDR_EQ12,
  ADDR_EQ13,
  ADDR_EQ14,
  ADDR_EQ15,
  ADDR_EQ16,
  ADDR_EQ17,
  ADDR_EQ18,
  ADDR_EQ19,
  ADDR_EQ20,
  ADDR_EQ21,
  ADDR_EQ22,
  ADDR_EQ23,
  ADDR_EQ24,
  ADDR_EQ25,
  ADDR_EQ26,
  ADDR_EQ27,
  ADDR_EQ28,
  ADDR_EQ29,
  ADDR_EQ30,
  ADDR_EQ31,
  BANK_TST2,
  BANK_TST3,
  BANK_TST4,
  BANK_TST5,
  BANK_TST6,
  BANK_TST7,
  BANK_TST8,
  BANK_TST9,
  BANK_TST10,
  BANK_TST11,
  BANK_TST12,
  BANK_TST13,
  BANK_TST14,
  BANK_TST15,
  LFSR_TST0,
  LFSR_TST1,
  LFSR_TST2,
  LFSR_TST3,
  LFSR_TST4,
  LFSR_TST5,
  LFSR_TST6,
  LFSR_TST7,
  REV_ADDR0,
  REV_ADDR1,
  MEM_BAND,
  INT_TST,
  FP_TST,
  DISK_TST,
  SWAP_TST,
  NUM_TESTS
};

enum{
  DISK_ONLY = 0x1,
  BAND_ONLY = 0x2,
  SWAP_ONLY = 0x4,
  FP_ONLY =      0x100L,
  INT_ONLY =     0x1000L,
  ADDR_EQ_ONLY = 0x10000L,
  BANK_ONLY =    0x20000L,
  PATTERN_ONLY = 0x40000L,
  RANDOM_ONLY =  0x80000L,
  WALKING_ONLY = 0x100000L,
  REV_ADDR_ONLY =0x200000L
};

enum{
  ADD_RR = 0,
  ADD_MR,
  MPY_RR,
  MPY_MR,
  DIV_RR,
  DIV_MR,
  DOT_MRR,
  DOT_MMR,
  NUM_FP_TESTS
};

enum{
  CACHED_SYNC=0,
  CACHED_NOSYNC,
  DIRECT_SYNC,
  DIRECT_NOSYNC,
  NUM_DISK_MODES
};

enum{
  DISK_READ=0,
  DISK_WRITE,
  NUM_DISK_OPS
};

typedef struct{
  char path[128][512];
  SInt32 num_paths;
} file_info;

typedef struct{
  SInt16 int16[MAX_SIZES][NUM_FP_TESTS];
  SInt32 int32[MAX_SIZES][NUM_FP_TESTS];
  SInt64 int64[MAX_SIZES][NUM_FP_TESTS];
  float fp[MAX_SIZES][NUM_FP_TESTS];
  double dbl[MAX_SIZES][NUM_FP_TESTS];
  SInt64 time_scale16[MAX_SIZES][NUM_FP_TESTS];
  SInt64 time_scale32[MAX_SIZES][NUM_FP_TESTS];
  SInt64 time_scale64[MAX_SIZES][NUM_FP_TESTS];
  SInt64 time_scale_fp[MAX_SIZES][NUM_FP_TESTS];
  SInt64 time_scale_dbl[MAX_SIZES][NUM_FP_TESTS];
} exp_data;

typedef struct{
  SInt64 runs;
  SInt64 errors;
  SInt64 size;
  double min;
  double max;
  double total;
  double total_time;
  char name[64];
} result;

typedef struct{
  result fixed[NUM_FIXED];
  result addreq[NUM_ADDREQ];
  result bank[NUM_BANK];
  result lfsr[NUM_LFSR];
  result walk[NUM_WALK];
  result rev_addr[NUM_REV_ADDR];
  result mem_bandwidth[2][MAX_SIZES];
  result swap[1];
  result int16_rslt[NUM_FP_TESTS][MAX_SIZES];
  result int32_rslt[NUM_FP_TESTS][MAX_SIZES];
  result int64_rslt[NUM_FP_TESTS][MAX_SIZES];
  result sgl_rslt[NUM_FP_TESTS][MAX_SIZES];
  result dbl_rslt[NUM_FP_TESTS][MAX_SIZES];
  result disk_rslt[NUM_DISK_OPS][MAX_DIRS][NUM_DISK_MODES][MAX_FILE_SIZES];
} result_summary;

typedef struct{
  Boolean rev_addr;
  Boolean lfsr;
  Boolean walk;
  Boolean addreq;
  Boolean bank;
  Boolean dbl;
  Boolean disk;
  Boolean fixed;
  Boolean int16;
  Boolean int32;
  Boolean int64;
  Boolean mem_bw;
  Boolean sgl;
  Boolean swap;
} run_summary;


typedef struct{
  SInt64 limit;
  SInt64 long_limit;
  UInt64 *seed;
  void *base;
  SInt32 errors;
  SInt32 num;
  SInt32 vnum;
  SInt32 num_threads;
  SInt32 test;
  SInt32 firsttime;
  exp_data *expected;
  file_info *finfo;
  result_summary *rslt_summary;
} thread_data;

typedef struct{
  SInt64 totalram;
  SInt64 freeram;
  SInt64 totalswap;
  SInt64 freeswap;
  SInt64 bufferram;
  SInt64 sharedram;
  SInt64 mem_unit;
} my_sysinfo;

typedef struct{
  time_t sec;
  suseconds_t usecs;
} time_def;

typedef struct {
  int tm_sec;         /* seconds */
  int tm_min;         /* minutes */
  int tm_hour;        /* hours */
  int tm_mday;        /* day of the month */
  int tm_mon;         /* month */
  int tm_year;        /* year */
  int tm_wday;        /* day of the week */
  int tm_yday;        /* day in the year */
  int tm_isdst;       /* daylight saving time */
} time_of_day;

typedef struct {
  UInt64 *select;
  double *minutes;
  SInt64 *dutycycle;
  file_info *finfo;
  Boolean *use_total;
  long *threads;
} config_args;

